@@ -32,13 +32,29 @@ const DefaultAPIServer = "http://localhost:8443"
 // IP address range from which subnet is allocated for per-node join switch
 const (
 	V4JoinSubnet = "100.64.0.0/16"
-	V6JoinSubnet = "fd98::/64"
+	V6JoinSubnet = "fd98::/48"
 )
 
+// Default IANA-assigned UDP port number for VXLAN
+const DefaultVXLANPort = 4789
+
 // The following are global config parameters that other modules may access directly
 var (
+	// Build information. Populated at build-time.
+	// commit ID used to build ovn-kubernetes
+	Commit = ""
+	// branch used to build ovn-kubernetes
+	Branch = ""
+	// ovn-kubernetes build user
+	BuildUser = ""
+	// ovn-kubernetes build date
+	BuildDate = ""
 	// ovn-kubernetes version, to be changed with every release
 	Version = "0.3.0"
+	// version of the go runtime used to compile ovn-kubernetes
+	GoVersion = runtime.Version()
+	// os and architecture used to build ovn-kubernetes
+	OSArch = fmt.Sprintf("%s %s", runtime.GOOS, runtime.GOARCH)
 
 	// ovn-kubernetes cni config file name
 	CNIConfFileName = "10-ovn-kubernetes.conf"
@@ -57,16 +73,18 @@ var (
 
 	// Logging holds logging-related parsed config file parameters and command-line overrides
 	Logging = LoggingConfig{
-		File:    "", // do not log to a file by default
-		CNIFile: "",
-		Level:   4,
+		File:              "", // do not log to a file by default
+		CNIFile:           "",
+		Level:             4,
+		LogFileMaxSize:    100, // Size in Megabytes
+		LogFileMaxBackups: 5,
+		LogFileMaxAge:     5, //days
 	}
 
 	// CNI holds CNI-related parsed config file parameters and command-line overrides
 	CNI = CNIConfig{
-		ConfDir:         "/etc/cni/net.d",
-		Plugin:          "ovn-k8s-cni-overlay",
-		WinHNSNetworkID: "",
+		ConfDir: "/etc/cni/net.d",
+		Plugin:  "ovn-k8s-cni-overlay",
 	}
 
 	// Kubernetes holds Kubernetes-related parsed config file parameters and command-line overrides
@@ -76,6 +94,11 @@ var (
 		OVNConfigNamespace: "ovn-kubernetes",
 	}
 
+	// OVNKubernetesFeatureConfig holds OVN-Kubernetes feature enhancement config file parameters and command-line overrides
+	OVNKubernetesFeature = OVNKubernetesFeatureConfig{
+		EnableEgressIP: true,
+	}
+
 	// OvnNorth holds northbound OVN database client and server authentication and location details
 	OvnNorth OvnAuthConfig
 
@@ -93,7 +116,9 @@ var (
 	}
 
 	// HybridOverlay holds hybrid overlay feature config options.
-	HybridOverlay HybridOverlayConfig
+	HybridOverlay = HybridOverlayConfig{
+		VXLANPort: DefaultVXLANPort,
+	}
 
 	// NbctlDaemon enables ovn-nbctl to run in daemon mode
 	NbctlDaemonMode bool
@@ -156,6 +181,13 @@ type LoggingConfig struct {
 	CNIFile string `gcfg:"cnilogfile"`
 	// Level is the logging verbosity level
 	Level int `gcfg:"loglevel"`
+	// LogFileMaxSize is the maximum size in bytes of the logfile
+	// before it gets rolled.
+	LogFileMaxSize int `gcfg:"logfile-maxsize"`
+	// LogFileMaxBackups represents the the maximum number of old log files to retain
+	LogFileMaxBackups int `gcfg:"logfile-maxbackups"`
+	// LogFileMaxAge represents the maximum number of days to retain old log files
+	LogFileMaxAge int `gcfg:"logfile-maxage"`
 }
 
 // CNIConfig holds CNI-related parsed config file parameters and command-line overrides
@@ -164,26 +196,30 @@ type CNIConfig struct {
 	ConfDir string `gcfg:"conf-dir"`
 	// Plugin specifies the name of the CNI plugin
 	Plugin string `gcfg:"plugin"`
-	// Windows ONLY, specifies the ID of the HNS Network to which the containers will be attached
-	WinHNSNetworkID string `gcfg:"win-hnsnetwork-id"`
 }
 
 // KubernetesConfig holds Kubernetes-related parsed config file parameters and command-line overrides
 type KubernetesConfig struct {
-	Kubeconfig           string `gcfg:"kubeconfig"`
-	CACert               string `gcfg:"cacert"`
-	APIServer            string `gcfg:"apiserver"`
-	Token                string `gcfg:"token"`
-	CompatServiceCIDR    string `gcfg:"service-cidr"`
-	RawServiceCIDRs      string `gcfg:"service-cidrs"`
-	ServiceCIDRs         []*net.IPNet
-	OVNConfigNamespace   string `gcfg:"ovn-config-namespace"`
-	MetricsBindAddress   string `gcfg:"metrics-bind-address"`
-	MetricsEnablePprof   bool   `gcfg:"metrics-enable-pprof"`
-	OVNEmptyLbEvents     bool   `gcfg:"ovn-empty-lb-events"`
-	PodIP                string `gcfg:"pod-ip"` // UNUSED
-	RawNoHostSubnetNodes string `gcfg:"no-hostsubnet-nodes"`
-	NoHostSubnetNodes    *metav1.LabelSelector
+	Kubeconfig            string `gcfg:"kubeconfig"`
+	CACert                string `gcfg:"cacert"`
+	APIServer             string `gcfg:"apiserver"`
+	Token                 string `gcfg:"token"`
+	CompatServiceCIDR     string `gcfg:"service-cidr"`
+	RawServiceCIDRs       string `gcfg:"service-cidrs"`
+	ServiceCIDRs          []*net.IPNet
+	OVNConfigNamespace    string `gcfg:"ovn-config-namespace"`
+	MetricsBindAddress    string `gcfg:"metrics-bind-address"`
+	OVNMetricsBindAddress string `gcfg:"ovn-metrics-bind-address"`
+	MetricsEnablePprof    bool   `gcfg:"metrics-enable-pprof"`
+	OVNEmptyLbEvents      bool   `gcfg:"ovn-empty-lb-events"`
+	PodIP                 string `gcfg:"pod-ip"` // UNUSED
+	RawNoHostSubnetNodes  string `gcfg:"no-hostsubnet-nodes"`
+	NoHostSubnetNodes     *metav1.LabelSelector
+}
+
+// OVNKubernetesFeatureConfig holds OVN-Kubernetes feature enhancement config file parameters and command-line overrides
+type OVNKubernetesFeatureConfig struct {
+	EnableEgressIP bool `gcfg:"enable-egress-ip"`
 }
 
 // GatewayMode holds the node gateway mode
@@ -210,17 +246,20 @@ type GatewayConfig struct {
 	VLANID uint `gcfg:"vlan-id"`
 	// NodeportEnable sets whether to provide Kubernetes NodePort service or not
 	NodeportEnable bool `gcfg:"nodeport"`
+	// DisableSNATMultipleGws sets whether to disable SNAT of egress traffic in namespaces annotated with routing-external-gws
+	DisableSNATMultipleGWs bool `gcfg:"disable-snat-multiple-gws"`
 }
 
 // OvnAuthConfig holds client authentication and location details for
 // an OVN database (either northbound or southbound)
 type OvnAuthConfig struct {
 	// e.g: "ssl:192.168.1.2:6641,ssl:192.168.1.2:6642"
-	Address string `gcfg:"address"`
-	PrivKey string `gcfg:"client-privkey"`
-	Cert    string `gcfg:"client-cert"`
-	CACert  string `gcfg:"client-cacert"`
-	Scheme  OvnDBScheme
+	Address        string `gcfg:"address"`
+	PrivKey        string `gcfg:"client-privkey"`
+	Cert           string `gcfg:"client-cert"`
+	CACert         string `gcfg:"client-cacert"`
+	CertCommonName string `gcfg:"cert-common-name"`
+	Scheme         OvnDBScheme
 
 	northbound bool
 	externalID string // ovn-nb or ovn-remote
@@ -247,6 +286,8 @@ type HybridOverlayConfig struct {
 	// ClusterSubnets holds parsed hybrid overlay cluster subnet entries and
 	// may be used outside the config module.
 	ClusterSubnets []CIDRNetworkEntry
+	// VXLANPort holds the VXLAN tunnel UDP port number.
+	VXLANPort uint `gcfg:"hybrid-overlay-vxlan-port"`
 }
 
 // OvnDBScheme describes the OVN database connection transport method
@@ -263,27 +304,29 @@ const (
 
 // Config is used to read the structured config file and to cache config in testcases
 type config struct {
-	Default       DefaultConfig
-	Logging       LoggingConfig
-	CNI           CNIConfig
-	Kubernetes    KubernetesConfig
-	OvnNorth      OvnAuthConfig
-	OvnSouth      OvnAuthConfig
-	Gateway       GatewayConfig
-	MasterHA      MasterHAConfig
-	HybridOverlay HybridOverlayConfig
+	Default              DefaultConfig
+	Logging              LoggingConfig
+	CNI                  CNIConfig
+	OVNKubernetesFeature OVNKubernetesFeatureConfig
+	Kubernetes           KubernetesConfig
+	OvnNorth             OvnAuthConfig
+	OvnSouth             OvnAuthConfig
+	Gateway              GatewayConfig
+	MasterHA             MasterHAConfig
+	HybridOverlay        HybridOverlayConfig
 }
 
 var (
-	savedDefault       DefaultConfig
-	savedLogging       LoggingConfig
-	savedCNI           CNIConfig
-	savedKubernetes    KubernetesConfig
-	savedOvnNorth      OvnAuthConfig
-	savedOvnSouth      OvnAuthConfig
-	savedGateway       GatewayConfig
-	savedMasterHA      MasterHAConfig
-	savedHybridOverlay HybridOverlayConfig
+	savedDefault              DefaultConfig
+	savedLogging              LoggingConfig
+	savedCNI                  CNIConfig
+	savedOVNKubernetesFeature OVNKubernetesFeatureConfig
+	savedKubernetes           KubernetesConfig
+	savedOvnNorth             OvnAuthConfig
+	savedOvnSouth             OvnAuthConfig
+	savedGateway              GatewayConfig
+	savedMasterHA             MasterHAConfig
+	savedHybridOverlay        HybridOverlayConfig
 	// legacy service-cluster-ip-range CLI option
 	serviceClusterIPRange string
 	// legacy cluster-subnet CLI option
@@ -299,14 +342,24 @@ func init() {
 	savedDefault = Default
 	savedLogging = Logging
 	savedCNI = CNI
+	savedOVNKubernetesFeature = OVNKubernetesFeature
 	savedKubernetes = Kubernetes
 	savedOvnNorth = OvnNorth
 	savedOvnSouth = OvnSouth
 	savedGateway = Gateway
 	savedMasterHA = MasterHA
 	savedHybridOverlay = HybridOverlay
+	cli.VersionPrinter = func(c *cli.Context) {
+		fmt.Printf("Version: %s\n", Version)
+		fmt.Printf("Git commit: %s\n", Commit)
+		fmt.Printf("Git branch: %s\n", Branch)
+		fmt.Printf("Go version: %s\n", GoVersion)
+		fmt.Printf("Build date: %s\n", BuildDate)
+		fmt.Printf("OS/Arch: %s\n", OSArch)
+	}
 	Flags = append(Flags, CommonFlags...)
 	Flags = append(Flags, CNIFlags...)
+	Flags = append(Flags, OVNK8sFeatureFlags...)
 	Flags = append(Flags, K8sFlags...)
 	Flags = append(Flags, OvnNBFlags...)
 	Flags = append(Flags, OvnSBFlags...)
@@ -321,6 +374,7 @@ func PrepareTestConfig() {
 	Default = savedDefault
 	Logging = savedLogging
 	CNI = savedCNI
+	OVNKubernetesFeature = savedOVNKubernetesFeature
 	Kubernetes = savedKubernetes
 	OvnNorth = savedOvnNorth
 	OvnSouth = savedOvnSouth
@@ -512,6 +566,25 @@ var CommonFlags = []cli.Flag{
 		Destination: &cliConfig.Logging.CNIFile,
 		Value:       "/var/log/ovn-kubernetes/ovn-k8s-cni-overlay.log",
 	},
+	// Logfile rotation parameters
+	&cli.IntFlag{
+		Name:        "logfile-maxsize",
+		Usage:       "Maximum size in bytes of the log file before it gets rolled",
+		Destination: &cliConfig.Logging.LogFileMaxSize,
+		Value:       Logging.LogFileMaxSize,
+	},
+	&cli.IntFlag{
+		Name:        "logfile-maxbackups",
+		Usage:       "Maximum number of old log files to retain",
+		Destination: &cliConfig.Logging.LogFileMaxBackups,
+		Value:       Logging.LogFileMaxBackups,
+	},
+	&cli.IntFlag{
+		Name:        "logfile-maxage",
+		Usage:       "Maximum number of days to retain old log files",
+		Destination: &cliConfig.Logging.LogFileMaxAge,
+		Value:       Logging.LogFileMaxAge,
+	},
 }
 
 // CNIFlags capture CNI-related options
@@ -529,10 +602,15 @@ var CNIFlags = []cli.Flag{
 		Destination: &cliConfig.CNI.Plugin,
 		Value:       CNI.Plugin,
 	},
-	&cli.StringFlag{
-		Name:        "win-hnsnetwork-id",
-		Usage:       "the ID of the HNS network to which containers will be attached (default: not set)",
-		Destination: &cliConfig.CNI.WinHNSNetworkID,
+}
+
+// OVNK8sFeatureFlags capture OVN-Kubernetes feature related options
+var OVNK8sFeatureFlags = []cli.Flag{
+	&cli.BoolFlag{
+		Name:        "enable-egress-ip",
+		Usage:       "Configure to use EgressIP CRD feature with ovn-kubernetes.",
+		Destination: &cliConfig.OVNKubernetesFeature.EnableEgressIP,
+		Value:       OVNKubernetesFeature.EnableEgressIP,
 	},
 }
 
@@ -586,9 +664,14 @@ var K8sFlags = []cli.Flag{
 	},
 	&cli.StringFlag{
 		Name:        "metrics-bind-address",
-		Usage:       "The IP address and port for the metrics server to serve on (set to 0.0.0.0 for all IPv4 interfaces)",
+		Usage:       "The IP address and port for the OVN K8s metrics server to serve on (set to 0.0.0.0 for all IPv4 interfaces)",
 		Destination: &cliConfig.Kubernetes.MetricsBindAddress,
 	},
+	&cli.StringFlag{
+		Name:        "ovn-metrics-bind-address",
+		Usage:       "The IP address and port for the OVN metrics server to serve on (set to 0.0.0.0 for all IPv4 interfaces)",
+		Destination: &cliConfig.Kubernetes.OVNMetricsBindAddress,
+	},
 	&cli.BoolFlag{
 		Name:        "metrics-enable-pprof",
 		Usage:       "If true, then also accept pprof requests on the metrics port.",
@@ -640,6 +723,14 @@ var OvnNBFlags = []cli.Flag{
 			"Default value for this setting is empty which defaults to use local unix socket.",
 		Destination: &cliConfig.OvnNorth.CACert,
 	},
+	&cli.StringFlag{
+		Name: "nb-cert-common-name",
+		Usage: "Common Name of the certificate used for TLS server certificate verification. " +
+			"In cases where the certificate doesn't have any SAN Extensions, this parameter " +
+			"should match the DNS(hostname) of the server. In case the certificate has a " +
+			"SAN extension, this parameter should match one of the SAN fields.",
+		Destination: &cliConfig.OvnNorth.CertCommonName,
+	},
 }
 
 //OvnSBFlags capture OVN southbound database options
@@ -669,6 +760,14 @@ var OvnSBFlags = []cli.Flag{
 			"Default value for this setting is empty which defaults to use local unix socket.",
 		Destination: &cliConfig.OvnSouth.CACert,
 	},
+	&cli.StringFlag{
+		Name: "sb-cert-common-name",
+		Usage: "Common Name of the certificate used for TLS server certificate verification. " +
+			"In cases where the certificate doesn't have any SAN Extensions, this parameter " +
+			"should match the DNS(hostname) of the server. In case the certificate has a " +
+			"SAN extension, this parameter should match one of the SAN fields.",
+		Destination: &cliConfig.OvnSouth.CertCommonName,
+	},
 }
 
 //OVNGatewayFlags capture L3 Gateway related flags
@@ -706,6 +805,11 @@ var OVNGatewayFlags = []cli.Flag{
 		Usage:       "Setup nodeport based ingress on gateways.",
 		Destination: &cliConfig.Gateway.NodeportEnable,
 	},
+	&cli.BoolFlag{
+		Name:        "disable-snat-multiple-gws",
+		Usage:       "Disable SNAT for egress traffic with multiple gateways.",
+		Destination: &cliConfig.Gateway.DisableSNATMultipleGWs,
+	},
 
 	// Deprecated CLI options
 	&cli.BoolFlag{
@@ -760,6 +864,12 @@ var HybridOverlayFlags = []cli.Flag{
 			"hostsubnetlength defines how many IP addresses are dedicated to each node.",
 		Destination: &cliConfig.HybridOverlay.RawClusterSubnets,
 	},
+	&cli.UintFlag{
+		Name:        "hybrid-overlay-vxlan-port",
+		Value:       HybridOverlay.VXLANPort,
+		Usage:       "The UDP port used by the VXLAN protocol for hybrid networks.",
+		Destination: &cliConfig.HybridOverlay.VXLANPort,
+	},
 }
 
 // Flags are general command-line flags. Apps should add these flags to their
@@ -771,6 +881,7 @@ var Flags []cli.Flag
 func GetFlags(customFlags []cli.Flag) []cli.Flag {
 	flags := CommonFlags
 	flags = append(flags, CNIFlags...)
+	flags = append(flags, OVNK8sFeatureFlags...)
 	flags = append(flags, K8sFlags...)
 	flags = append(flags, OvnNBFlags...)
 	flags = append(flags, OvnSBFlags...)
@@ -829,7 +940,7 @@ func getOVSExternalID(exec kexec.Interface, name string) string {
 		".",
 		"external_ids:"+name)
 	if err != nil {
-		klog.V(5).Infof("failed to get OVS external_id %s: %v\n\t%s", name, err, out)
+		klog.V(5).Infof("Failed to get OVS external_id %s: %v\n\t%s", name, err, out)
 		return ""
 	}
 	return out
@@ -842,7 +953,7 @@ func setOVSExternalID(exec kexec.Interface, key, value string) error {
 		".",
 		fmt.Sprintf("external_ids:%s=%s", key, value))
 	if err != nil {
-		return fmt.Errorf("Error setting OVS external ID '%s=%s': %v\n  %q", key, value, err, out)
+		return fmt.Errorf("error setting OVS external ID '%s=%s': %v\n  %q", key, value, err, out)
 	}
 	return nil
 }
@@ -993,9 +1104,22 @@ func buildGatewayConfig(ctx *cli.Context, cli, file *config) error {
 		if Gateway.NextHop != "" {
 			return fmt.Errorf("gateway next-hop option %q not allowed when gateway is disabled", Gateway.NextHop)
 		}
-		if Gateway.VLANID != 0 {
-			return fmt.Errorf("gateway VLAN ID option '%d' not allowed when gateway is disabled", Gateway.VLANID)
-		}
+	}
+
+	if Gateway.Mode != GatewayModeShared && Gateway.VLANID != 0 {
+		return fmt.Errorf("gateway VLAN ID option: %d is supported only in shared gateway mode", Gateway.VLANID)
+	}
+	return nil
+}
+
+func buildOVNKubernetesFeatureConfig(ctx *cli.Context, cli, file *config) error {
+	// Copy config file values over default values
+	if err := overrideFields(&OVNKubernetesFeature, &file.OVNKubernetesFeature, &savedOVNKubernetesFeature); err != nil {
+		return err
+	}
+	// And CLI overrides over config file and default values
+	if err := overrideFields(&OVNKubernetesFeature, &cli.OVNKubernetesFeature, &savedOVNKubernetesFeature); err != nil {
+		return err
 	}
 	return nil
 }
@@ -1012,13 +1136,13 @@ func buildMasterHAConfig(ctx *cli.Context, cli, file *config) error {
 	}
 
 	if MasterHA.ElectionLeaseDuration <= MasterHA.ElectionRenewDeadline {
-		return fmt.Errorf("Invalid HA election lease duration '%d'. "+
+		return fmt.Errorf("invalid HA election lease duration '%d'. "+
 			"It should be greater than HA election renew deadline '%d'",
 			MasterHA.ElectionLeaseDuration, MasterHA.ElectionRenewDeadline)
 	}
 
 	if MasterHA.ElectionRenewDeadline <= MasterHA.ElectionRetryPeriod {
-		return fmt.Errorf("Invalid HA election renew deadline duration '%d'. "+
+		return fmt.Errorf("invalid HA election renew deadline duration '%d'. "+
 			"It should be greater than HA election retry period '%d'",
 			MasterHA.ElectionRenewDeadline, MasterHA.ElectionRetryPeriod)
 	}
@@ -1047,6 +1171,9 @@ func buildHybridOverlayConfig(ctx *cli.Context, cli, file *config, allSubnets *c
 				allSubnets.append(configSubnetHybrid, subnet.CIDR)
 			}
 		}
+		if HybridOverlay.VXLANPort > 65535 {
+			return fmt.Errorf("hybrid overlay vxlan port is invalid. The port cannot be larger than 65535")
+		}
 	}
 
 	return nil
@@ -1089,14 +1216,7 @@ func getConfigFilePath(ctx *cli.Context) (string, bool) {
 	if configFile != "" {
 		return configFile, false
 	}
-
-	// Linux default
-	if runtime.GOOS != "windows" {
-		return filepath.Join("/etc", "openvswitch", "ovn_k8s.conf"), true
-	}
-
-	// Windows default
-	return filepath.Join(os.Getenv("OVS_SYSCONFDIR"), "ovn_k8s.conf"), true
+	return "/etc/openvswitch/ovn_k8s.conf", true
 }
 
 // InitConfig reads the config file and common command-line options and
@@ -1124,15 +1244,16 @@ func initConfigWithPath(ctx *cli.Context, exec kexec.Interface, saPath string, d
 	var err error
 	// initialize cfg with default values, allow file read to override
 	cfg := config{
-		Default:       savedDefault,
-		Logging:       savedLogging,
-		CNI:           savedCNI,
-		Kubernetes:    savedKubernetes,
-		OvnNorth:      savedOvnNorth,
-		OvnSouth:      savedOvnSouth,
-		Gateway:       savedGateway,
-		MasterHA:      savedMasterHA,
-		HybridOverlay: savedHybridOverlay,
+		Default:              savedDefault,
+		Logging:              savedLogging,
+		CNI:                  savedCNI,
+		OVNKubernetesFeature: savedOVNKubernetesFeature,
+		Kubernetes:           savedKubernetes,
+		OvnNorth:             savedOvnNorth,
+		OvnSouth:             savedOvnSouth,
+		Gateway:              savedGateway,
+		MasterHA:             savedMasterHA,
+		HybridOverlay:        savedHybridOverlay,
 	}
 
 	allSubnets := newConfigSubnets()
@@ -1197,9 +1318,9 @@ func initConfigWithPath(ctx *cli.Context, exec kexec.Interface, saPath string, d
 		}
 		klog.SetOutput(&lumberjack.Logger{
 			Filename:   Logging.File,
-			MaxSize:    100, // megabytes
-			MaxBackups: 10,
-			MaxAge:     30, // days
+			MaxSize:    Logging.LogFileMaxSize, // megabytes
+			MaxBackups: Logging.LogFileMaxBackups,
+			MaxAge:     Logging.LogFileMaxAge, // days
 			Compress:   true,
 		})
 	}
@@ -1212,6 +1333,10 @@ func initConfigWithPath(ctx *cli.Context, exec kexec.Interface, saPath string, d
 		return "", err
 	}
 
+	if err = buildOVNKubernetesFeatureConfig(ctx, &cliConfig, &cfg); err != nil {
+		return "", err
+	}
+
 	if err = buildGatewayConfig(ctx, &cliConfig, &cfg); err != nil {
 		return "", err
 	}
@@ -1279,13 +1404,13 @@ func parseAddress(urlString string) (string, OvnDBScheme, error) {
 	for _, ovnAddress := range strings.Split(urlString, ",") {
 		splits := strings.SplitN(ovnAddress, ":", 2)
 		if len(splits) != 2 {
-			return "", "", fmt.Errorf("Failed to parse OVN address %s", urlString)
+			return "", "", fmt.Errorf("failed to parse OVN address %s", urlString)
 		}
 
 		if scheme == "" {
 			scheme = splits[0]
 		} else if scheme != splits[0] {
-			return "", "", fmt.Errorf("Invalid protocols in OVN address %s",
+			return "", "", fmt.Errorf("invalid protocols in OVN address %s",
 				urlString)
 		}
 
@@ -1372,8 +1497,8 @@ func buildOvnAuth(exec kexec.Interface, northbound bool, cliAuth, confAuth *OvnA
 
 	switch {
 	case auth.Scheme == OvnDBSchemeSSL:
-		if auth.PrivKey == "" || auth.Cert == "" || auth.CACert == "" {
-			return nil, fmt.Errorf("must specify private key, certificate, and CA certificate for 'ssl' scheme")
+		if auth.PrivKey == "" || auth.Cert == "" || auth.CACert == "" || auth.CertCommonName == "" {
+			return nil, fmt.Errorf("must specify private key, certificate, CA certificate, and common name used in the certificate for 'ssl' scheme")
 		}
 	case auth.Scheme == OvnDBSchemeTCP:
 		if auth.PrivKey != "" || auth.Cert != "" || auth.CACert != "" {
@@ -1407,7 +1532,7 @@ func (a *OvnAuthConfig) ensureCACert() error {
 	args = append(args, "list", "nb_global")
 	_, _ = rawExec(a.exec, "ovn-nbctl", args...)
 	if _, err := os.Stat(a.CACert); os.IsNotExist(err) {
-		klog.Warningf("bootstrapping %s CA certificate failed", a.CACert)
+		klog.Warningf("Bootstrapping %s CA certificate failed", a.CACert)
 	}
 	return nil
 }
