@@ -2,17 +2,18 @@ package ovn
 
 import (
 	"context"
-	"encoding/json"
 	"fmt"
 	"net"
 	"os"
 	"reflect"
 	"strings"
+	"sync"
 	"time"
 
 	kapi "k8s.io/api/core/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	utilruntime "k8s.io/apimachinery/pkg/util/runtime"
+	utilwait "k8s.io/apimachinery/pkg/util/wait"
 	"k8s.io/client-go/kubernetes"
 	"k8s.io/client-go/tools/leaderelection"
 	"k8s.io/client-go/tools/leaderelection/resourcelock"
@@ -20,20 +21,41 @@ import (
 	"k8s.io/klog"
 	utilnet "k8s.io/utils/net"
 
-	homaster "github.com/ovn-org/ovn-kubernetes/go-controller/hybrid-overlay/pkg/controller"
+	goovn "github.com/ebay/go-ovn"
+	hocontroller "github.com/ovn-org/ovn-kubernetes/go-controller/hybrid-overlay/pkg/controller"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/config"
+	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/informer"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/metrics"
+	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/ovn/ipallocator"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/util"
 )
 
 const (
 	// OvnServiceIdledAt is a constant string representing the Service annotation key
 	// whose value indicates the time stamp in RFC3339 format when a Service was idled
-	OvnServiceIdledAt = "k8s.ovn.org/idled-at"
+	OvnServiceIdledAt              = "k8s.ovn.org/idled-at"
+	OvnNodeAnnotationRetryInterval = 100 * time.Millisecond
+	OvnNodeAnnotationRetryTimeout  = 1 * time.Second
 )
 
+type ovnkubeMasterLeaderMetrics struct{}
+
+func (ovnkubeMasterLeaderMetrics) On(string) {
+	metrics.MetricMasterLeader.Set(1)
+}
+
+func (ovnkubeMasterLeaderMetrics) Off(string) {
+	metrics.MetricMasterLeader.Set(0)
+}
+
+type ovnkubeMasterLeaderMetricsProvider struct{}
+
+func (_ ovnkubeMasterLeaderMetricsProvider) NewLeaderMetric() leaderelection.SwitchMetric {
+	return ovnkubeMasterLeaderMetrics{}
+}
+
 // Start waits until this process is the leader before starting master functions
-func (oc *Controller) Start(kClient kubernetes.Interface, nodeName string) error {
+func (oc *Controller) Start(kClient kubernetes.Interface, nodeName string, wg *sync.WaitGroup) error {
 	// Set up leader election process first
 	rl, err := resourcelock.New(
 		resourcelock.ConfigMapsResourceLock,
@@ -54,17 +76,20 @@ func (oc *Controller) Start(kClient kubernetes.Interface, nodeName string) error
 		RetryPeriod:   time.Duration(config.MasterHA.ElectionRetryPeriod) * time.Second,
 		Callbacks: leaderelection.LeaderCallbacks{
 			OnStartedLeading: func(ctx context.Context) {
-				klog.Infof("won leader election; in active mode")
+				klog.Infof("Won leader election; in active mode")
 				// run the cluster controller to init the master
 				start := time.Now()
 				defer func() {
 					end := time.Since(start)
 					metrics.MetricMasterReadyDuration.Set(end.Seconds())
 				}()
+				// run the End-to-end timestamp metric updater only on the
+				// active master node.
+				metrics.StartE2ETimeStampMetricUpdater(oc.stopChan, oc.ovnNBClient)
 				if err := oc.StartClusterMaster(nodeName); err != nil {
 					panic(err.Error())
 				}
-				if err := oc.Run(); err != nil {
+				if err := oc.Run(wg); err != nil {
 					panic(err.Error())
 				}
 			},
@@ -74,17 +99,18 @@ func (oc *Controller) Start(kClient kubernetes.Interface, nodeName string) error
 				// we need to handle the transition properly like clearing
 				// the cache. It is better to exit for now.
 				// kube will restart and this will become a follower.
-				klog.Infof("no longer leader; exiting")
+				klog.Infof("No longer leader; exiting")
 				os.Exit(1)
 			},
 			OnNewLeader: func(newLeaderName string) {
 				if newLeaderName != nodeName {
-					klog.Infof("lost the election to %s; in standby mode", newLeaderName)
+					klog.Infof("Lost the election to %s; in standby mode", newLeaderName)
 				}
 			},
 		},
 	}
 
+	leaderelection.SetProvider(ovnkubeMasterLeaderMetricsProvider{})
 	leaderElector, err := leaderelection.NewLeaderElector(lec)
 	if err != nil {
 		return err
@@ -105,15 +131,30 @@ func (oc *Controller) Start(kClient kubernetes.Interface, nodeName string) error
 //  If true, then either quit or perform a complete reconfiguration of the cluster (recreate switches/routers with new subnet values)
 func (oc *Controller) StartClusterMaster(masterNodeName string) error {
 	// The gateway router need to be connected to the distributed router via a per-node join switch.
-	// We need a subnet allocator that allocates subnet for this per-node join switch. Use the 100.64.0.0/16
-	// or fd98::/64 network range with host bits set to 3. The allocator will start allocating subnet that has upto 6
-	// host IPs)
-	joinSubnet := config.V4JoinSubnet
+	// We need a subnet allocator that allocates subnet for this per-node join switch.
+	if config.IPv4Mode {
+		// Use 100.64.0.0/16 with /29 subnets, allowing 8192 nodes with 6 IPs on each. (The join
+		// switches currently only have 2 IPs on them, so this leaves some room for expansion.)
+		_, joinSubnetCIDR, _ := net.ParseCIDR(config.V4JoinSubnet)
+		_ = oc.joinSubnetAllocator.AddNetworkRange(joinSubnetCIDR, 29)
+		// initialize the subnet required for DNAT and SNAT ip for the shared gateway mode
+		_, nodeLocalNatSubnetCIDR, _ := net.ParseCIDR(util.V4NodeLocalNatSubnet)
+		oc.nodeLocalNatIPv4Allocator, _ = ipallocator.NewCIDRRange(nodeLocalNatSubnetCIDR)
+		// set aside the first two IPs for the nextHop on the host and for distributed gateway port
+		_ = oc.nodeLocalNatIPv4Allocator.Allocate(net.ParseIP(util.V4NodeLocalNatSubnetNextHop))
+		_ = oc.nodeLocalNatIPv4Allocator.Allocate(net.ParseIP(util.V4NodeLocalDistributedGwPortIP))
+	}
 	if config.IPv6Mode {
-		joinSubnet = config.V6JoinSubnet
+		// Use fd98::/48 with /64 subnets
+		_, joinSubnetCIDR, _ := net.ParseCIDR(config.V6JoinSubnet)
+		_ = oc.joinSubnetAllocator.AddNetworkRange(joinSubnetCIDR, 64)
+		// initialize the subnet required for DNAT and SNAT ip for the shared gateway mode
+		_, nodeLocalNatSubnetCIDR, _ := net.ParseCIDR(util.V6NodeLocalNatSubnet)
+		oc.nodeLocalNatIPv6Allocator, _ = ipallocator.NewCIDRRange(nodeLocalNatSubnetCIDR)
+		// set aside the first two IPs for the nextHop on the host and for distributed gateway port
+		_ = oc.nodeLocalNatIPv6Allocator.Allocate(net.ParseIP(util.V6NodeLocalNatSubnetNextHop))
+		_ = oc.nodeLocalNatIPv6Allocator.Allocate(net.ParseIP(util.V6NodeLocalDistributedGwPortIP))
 	}
-	_, joinSubnetCIDR, _ := net.ParseCIDR(joinSubnet)
-	_ = oc.joinSubnetAllocator.AddNetworkRange(joinSubnetCIDR, 3)
 
 	existingNodes, err := oc.kube.GetNodes()
 	if err != nil {
@@ -121,7 +162,7 @@ func (oc *Controller) StartClusterMaster(masterNodeName string) error {
 		return err
 	}
 	for _, clusterEntry := range config.Default.ClusterSubnets {
-		err := oc.masterSubnetAllocator.AddNetworkRange(clusterEntry.CIDR, clusterEntry.HostBits())
+		err := oc.masterSubnetAllocator.AddNetworkRange(clusterEntry.CIDR, clusterEntry.HostSubnetLength)
 		if err != nil {
 			return err
 		}
@@ -141,10 +182,22 @@ func (oc *Controller) StartClusterMaster(masterNodeName string) error {
 				utilruntime.HandleError(err)
 			}
 		}
+		nodeLocalNatIPs, _ := util.ParseNodeLocalNatIPAnnotation(&node)
+		for _, nodeLocalNatIP := range nodeLocalNatIPs {
+			var err error
+			if utilnet.IsIPv6(nodeLocalNatIP) {
+				err = oc.nodeLocalNatIPv6Allocator.Allocate(nodeLocalNatIP)
+			} else {
+				err = oc.nodeLocalNatIPv4Allocator.Allocate(nodeLocalNatIP)
+			}
+			if err != nil {
+				utilruntime.HandleError(err)
+			}
+		}
 	}
 
 	if _, _, err := util.RunOVNNbctl("--columns=_uuid", "list", "port_group"); err != nil {
-		klog.Fatal("ovn version too old; does not support port groups")
+		klog.Fatal("OVN version too old; does not support port groups")
 	}
 
 	if oc.multicastSupport {
@@ -153,8 +206,8 @@ func (oc *Controller) StartClusterMaster(masterNodeName string) error {
 				"Disabling Multicast Support")
 			oc.multicastSupport = false
 		}
-		if config.IPv6Mode {
-			klog.Warningf("Multicast support enabled, but can not be used along with IPv6. Disabling Multicast Support")
+		if !config.IPv4Mode {
+			klog.Warningf("Multicast support enabled, but can not be used with single-stack IPv6. Disabling Multicast Support")
 			oc.multicastSupport = false
 		}
 	}
@@ -165,9 +218,18 @@ func (oc *Controller) StartClusterMaster(masterNodeName string) error {
 	}
 
 	if config.HybridOverlay.Enabled {
-		if err := homaster.StartMaster(oc.kube, oc.watchFactory); err != nil {
-			klog.Errorf("Failed to set up hybrid overlay master: %v", err)
-			return err
+		factory := oc.watchFactory.GetFactory()
+		oc.hoMaster, err = hocontroller.NewMaster(
+			oc.kube,
+			factory.Core().V1().Nodes().Informer(),
+			factory.Core().V1().Namespaces().Informer(),
+			factory.Core().V1().Pods().Informer(),
+			oc.ovnNBClient,
+			oc.ovnSBClient,
+			informer.NewDefaultEventHandler,
+		)
+		if err != nil {
+			return fmt.Errorf("failed to set up hybrid overlay master: %v", err)
 		}
 	}
 
@@ -185,6 +247,10 @@ func (oc *Controller) SetupMaster(masterNodeName string) error {
 		return err
 	}
 
+	if err := addDistributedGWPort(); err != nil {
+		return err
+	}
+
 	// Determine SCTP support
 	oc.SCTPSupport, err = util.DetectSCTPSupport()
 	if err != nil {
@@ -220,40 +286,40 @@ func (oc *Controller) SetupMaster(masterNodeName string) error {
 	// Create 3 load-balancers for east-west traffic for UDP, TCP, SCTP
 	oc.TCPLoadBalancerUUID, stderr, err = util.RunOVNNbctl("--data=bare", "--no-heading", "--columns=_uuid", "find", "load_balancer", "external_ids:k8s-cluster-lb-tcp=yes")
 	if err != nil {
-		klog.Errorf("Failed to get tcp load-balancer, stderr: %q, error: %v", stderr, err)
+		klog.Errorf("Failed to get tcp load balancer, stderr: %q, error: %v", stderr, err)
 		return err
 	}
 
 	if oc.TCPLoadBalancerUUID == "" {
 		oc.TCPLoadBalancerUUID, stderr, err = util.RunOVNNbctl("--", "create", "load_balancer", "external_ids:k8s-cluster-lb-tcp=yes", "protocol=tcp")
 		if err != nil {
-			klog.Errorf("Failed to create tcp load-balancer, stdout: %q, stderr: %q, error: %v", stdout, stderr, err)
+			klog.Errorf("Failed to create tcp load balancer, stdout: %q, stderr: %q, error: %v", stdout, stderr, err)
 			return err
 		}
 	}
 
 	oc.UDPLoadBalancerUUID, stderr, err = util.RunOVNNbctl("--data=bare", "--no-heading", "--columns=_uuid", "find", "load_balancer", "external_ids:k8s-cluster-lb-udp=yes")
 	if err != nil {
-		klog.Errorf("Failed to get udp load-balancer, stderr: %q, error: %v", stderr, err)
+		klog.Errorf("Failed to get udp load balancer, stderr: %q, error: %v", stderr, err)
 		return err
 	}
 	if oc.UDPLoadBalancerUUID == "" {
 		oc.UDPLoadBalancerUUID, stderr, err = util.RunOVNNbctl("--", "create", "load_balancer", "external_ids:k8s-cluster-lb-udp=yes", "protocol=udp")
 		if err != nil {
-			klog.Errorf("Failed to create udp load-balancer, stdout: %q, stderr: %q, error: %v", stdout, stderr, err)
+			klog.Errorf("Failed to create udp load balancer, stdout: %q, stderr: %q, error: %v", stdout, stderr, err)
 			return err
 		}
 	}
 
 	oc.SCTPLoadBalancerUUID, stderr, err = util.RunOVNNbctl("--data=bare", "--no-heading", "--columns=_uuid", "find", "load_balancer", "external_ids:k8s-cluster-lb-sctp=yes")
 	if err != nil {
-		klog.Errorf("Failed to get sctp load-balancer, stderr: %q, error: %v", stderr, err)
+		klog.Errorf("Failed to get sctp load balancer, stderr: %q, error: %v", stderr, err)
 		return err
 	}
 	if oc.SCTPLoadBalancerUUID == "" && oc.SCTPSupport {
 		oc.SCTPLoadBalancerUUID, stderr, err = util.RunOVNNbctl("--", "create", "load_balancer", "external_ids:k8s-cluster-lb-sctp=yes", "protocol=sctp")
 		if err != nil {
-			klog.Errorf("Failed to create sctp load-balancer, stdout: %q, stderr: %q, error: %v", stdout, stderr, err)
+			klog.Errorf("Failed to create sctp load balancer, stdout: %q, stderr: %q, error: %v", stdout, stderr, err)
 			return err
 		}
 	}
@@ -283,7 +349,7 @@ func (oc *Controller) allocateJoinSubnet(node *kapi.Node) ([]*net.IPNet, error)
 	// Allocate a new network for the join switch
 	joinSubnets, err = oc.joinSubnetAllocator.AllocateNetworks()
 	if err != nil {
-		return nil, fmt.Errorf("Error allocating subnet for join switch for node %s: %v", node.Name, err)
+		return nil, fmt.Errorf("error allocating subnet for join switch for node %s: %v", node.Name, err)
 	}
 
 	defer func() {
@@ -308,7 +374,7 @@ func (oc *Controller) allocateJoinSubnet(node *kapi.Node) ([]*net.IPNet, error)
 func (oc *Controller) deleteNodeJoinSubnet(nodeName string, subnet *net.IPNet) error {
 	err := oc.joinSubnetAllocator.ReleaseNetwork(subnet)
 	if err != nil {
-		return fmt.Errorf("Error deleting join subnet %v for node %q: %s", subnet, nodeName, err)
+		return fmt.Errorf("error deleting join subnet %v for node %q: %s", subnet, nodeName, err)
 	}
 	klog.Infof("Deleted JoinSubnet %v for node %s", subnet, nodeName)
 	return nil
@@ -320,16 +386,6 @@ func (oc *Controller) syncNodeManagementPort(node *kapi.Node, hostSubnets []*net
 		return err
 	}
 
-	if macAddress == nil {
-		// When macAddress was removed, delete the switch port
-		stdout, stderr, err := util.RunOVNNbctl("--", "--if-exists", "lsp-del", "k8s-"+node.Name)
-		if err != nil {
-			klog.Errorf("Failed to delete logical port to switch, stdout: %q, stderr: %q, error: %v", stdout, stderr, err)
-		}
-
-		return nil
-	}
-
 	if hostSubnets == nil {
 		hostSubnets, err = util.ParseNodeHostSubnetAnnotation(node)
 		if err != nil {
@@ -350,12 +406,23 @@ func (oc *Controller) syncNodeManagementPort(node *kapi.Node, hostSubnets []*net
 		if !utilnet.IsIPv6CIDR(hostSubnet) {
 			v4Subnet = hostSubnet
 		}
+
+		if config.Gateway.Mode == config.GatewayModeLocal {
+			stdout, stderr, err := util.RunOVNNbctl("--may-exist",
+				"--policy=src-ip", "lr-route-add", ovnClusterRouter,
+				hostSubnet.String(), mgmtIfAddr.IP.String())
+			if err != nil {
+				return fmt.Errorf("failed to add source IP address based "+
+					"routes in distributed router %s, stdout: %q, "+
+					"stderr: %q, error: %v", ovnClusterRouter, stdout, stderr, err)
+			}
+		}
 	}
 
 	// Create this node's management logical port on the node switch
 	stdout, stderr, err := util.RunOVNNbctl(
-		"--", "--may-exist", "lsp-add", node.Name, "k8s-"+node.Name,
-		"--", "lsp-set-addresses", "k8s-"+node.Name, addresses)
+		"--", "--may-exist", "lsp-add", node.Name, util.K8sPrefix+node.Name,
+		"--", "lsp-set-addresses", util.K8sPrefix+node.Name, addresses)
 	if err != nil {
 		klog.Errorf("Failed to add logical port to switch, stdout: %q, stderr: %q, error: %v", stdout, stderr, err)
 		return err
@@ -370,7 +437,8 @@ func (oc *Controller) syncNodeManagementPort(node *kapi.Node, hostSubnets []*net
 	return nil
 }
 
-func (oc *Controller) syncGatewayLogicalNetwork(node *kapi.Node, l3GatewayConfig *util.L3GatewayConfig, hostSubnets []*net.IPNet) error {
+func (oc *Controller) syncGatewayLogicalNetwork(node *kapi.Node, l3GatewayConfig *util.L3GatewayConfig,
+	hostSubnets []*net.IPNet) error {
 	var err error
 	var clusterSubnets []*net.IPNet
 	for _, clusterSubnet := range config.Default.ClusterSubnets {
@@ -383,31 +451,62 @@ func (oc *Controller) syncGatewayLogicalNetwork(node *kapi.Node, l3GatewayConfig
 		return err
 	}
 
-	err = gatewayInit(node.Name, clusterSubnets, hostSubnets, joinSubnets, l3GatewayConfig, oc.SCTPSupport)
-	if err != nil {
-		return fmt.Errorf("failed to init shared interface gateway: %v", err)
+	// OCP HACK
+	// GatewayModeLocal is only used if Local mode is specified and None shared gateway bridge is specified
+	// This is to allow local gateway mode without having to configure/use the shared gateway bridge
+	// See https://github.com/openshift/ovn-kubernetes/pull/281
+	if l3GatewayConfig.Mode == config.GatewayModeLocal {
+		err = gatewayInitMinimal(node.Name, l3GatewayConfig, oc.SCTPSupport)
+		if err != nil {
+			return fmt.Errorf("failed to init local gateway with no OVS bridge: %v", err)
+		}
+		// END OCP HACK
+	} else {
+		err = gatewayInit(node.Name, clusterSubnets, hostSubnets, joinSubnets, l3GatewayConfig, oc.SCTPSupport)
+		if err != nil {
+			return fmt.Errorf("failed to init shared interface gateway: %v", err)
+		}
 	}
 
-	if l3GatewayConfig.Mode == config.GatewayModeShared {
-		// Add static routes to OVN Cluster Router to enable pods on this Node to
-		// reach the host IP
-		err = addStaticRoutesToHost(node, l3GatewayConfig.IPAddresses)
+	// in the case of shared gateway mode, we need to setup
+	// 1. two policy based routes to steer traffic to the k8s node IP
+	// 	  - from the management port via the node_local_switch's localnet port
+	//    - from the hostsubnet via management port
+	// 2. a dnat_and_snat nat entry to SNAT the traffic from the management port
+	subnets, err := util.ParseNodeHostSubnetAnnotation(node)
+	if err != nil {
+		return fmt.Errorf("failed to get host subnets for %s: %v", node.Name, err)
+	}
+	mpMAC, err := util.ParseNodeManagementPortMACAddress(node)
+	if err != nil {
+		return err
+	}
+	for _, subnet := range subnets {
+		hostIfAddr := util.GetNodeManagementIfAddr(subnet)
+		l3GatewayConfigIP, err := util.MatchIPNetFamily(utilnet.IsIPv6(hostIfAddr.IP), l3GatewayConfig.IPAddresses)
 		if err != nil {
 			return err
 		}
+		if err := addPolicyBasedRoutes(node.Name, hostIfAddr.IP.String(), l3GatewayConfigIP); err != nil {
+			return err
+		}
+
+		if err := oc.addNodeLocalNatEntries(node, mpMAC.String(), hostIfAddr); err != nil {
+			return err
+		}
 	}
 
 	if l3GatewayConfig.NodePortEnable {
 		err = oc.handleNodePortLB(node)
 	} else {
 		// nodePort disabled, delete gateway load balancers for this node.
-		physicalGateway := "GR_" + node.Name
+		gatewayRouter := "GR_" + node.Name
 		for _, proto := range []kapi.Protocol{kapi.ProtocolTCP, kapi.ProtocolUDP, kapi.ProtocolSCTP} {
-			lbUUID, _ := oc.getGatewayLoadBalancer(physicalGateway, proto)
+			lbUUID, _ := oc.getGatewayLoadBalancer(gatewayRouter, proto)
 			if lbUUID != "" {
 				_, _, err := util.RunOVNNbctl("--if-exists", "destroy", "load_balancer", lbUUID)
 				if err != nil {
-					klog.Errorf("failed to destroy %s load balancer for gateway %s: %v", proto, physicalGateway, err)
+					klog.Errorf("Failed to destroy %s load balancer for gateway %s: %v", proto, gatewayRouter, err)
 				}
 			}
 		}
@@ -416,46 +515,6 @@ func (oc *Controller) syncGatewayLogicalNetwork(node *kapi.Node, l3GatewayConfig
 	return err
 }
 
-func hostAddrForSubnet(hostIfAddrs []*net.IPNet, subnet *net.IPNet) (string, error) {
-	isIPv6 := utilnet.IsIPv6CIDR(subnet)
-	for _, ifaddr := range hostIfAddrs {
-		if utilnet.IsIPv6CIDR(ifaddr) == isIPv6 {
-			if isIPv6 {
-				return ifaddr.IP.String() + "/128", nil
-			} else {
-				return ifaddr.IP.String() + "/32", nil
-			}
-		}
-	}
-	if isIPv6 {
-		return "", fmt.Errorf("no IPv6 host address available")
-	} else {
-		return "", fmt.Errorf("no IPv4 host address available")
-	}
-}
-
-func addStaticRoutesToHost(node *kapi.Node, hostIfAddrs []*net.IPNet) error {
-	subnets, err := util.ParseNodeHostSubnetAnnotation(node)
-	if err != nil {
-		return fmt.Errorf("failed to get host subnets for %s: %v", node.Name, err)
-	}
-
-	for _, subnet := range subnets {
-		hostAddr, err := hostAddrForSubnet(hostIfAddrs, subnet)
-		if err != nil {
-			return fmt.Errorf("cannot configure static route for %s: %v", subnet.String(), err)
-		}
-		nextHop := util.GetNodeManagementIfAddr(subnet).IP.String()
-		_, stderr, err := util.RunOVNNbctl("--may-exist", "lr-route-add", ovnClusterRouter, hostAddr, nextHop)
-		if err != nil {
-			return fmt.Errorf("failed to add static route '%s via %s' for host %q on %s "+
-				"stderr: %q, error: %v", hostAddr, nextHop, node.Name, ovnClusterRouter, stderr, err)
-		}
-	}
-
-	return nil
-}
-
 func (oc *Controller) ensureNodeLogicalNetwork(nodeName string, hostSubnets []*net.IPNet) error {
 	// logical router port MAC is based on IPv4 subnet if there is one, else IPv6
 	var nodeLRPMAC net.HardwareAddr
@@ -468,8 +527,8 @@ func (oc *Controller) ensureNodeLogicalNetwork(nodeName string, hostSubnets []*n
 	}
 
 	lrpArgs := []string{
-		"--if-exists", "lrp-del", "rtos-" + nodeName,
-		"--", "lrp-add", ovnClusterRouter, "rtos-" + nodeName,
+		"--if-exists", "lrp-del", routerToSwitchPrefix + nodeName,
+		"--", "lrp-add", ovnClusterRouter, routerToSwitchPrefix + nodeName,
 		nodeLRPMAC.String(),
 	}
 
@@ -554,8 +613,9 @@ func (oc *Controller) ensureNodeLogicalNetwork(nodeName string, hostSubnets []*n
 	}
 
 	// Connect the switch to the router.
-	stdout, stderr, err = util.RunOVNNbctl("--", "--may-exist", "lsp-add", nodeName, "stor-"+nodeName,
-		"--", "set", "logical_switch_port", "stor-"+nodeName, "type=router", "options:router-port=rtos-"+nodeName, "addresses="+"\""+nodeLRPMAC.String()+"\"")
+	stdout, stderr, err = util.RunOVNNbctl("--", "--may-exist", "lsp-add", nodeName, switchToRouterPrefix+nodeName,
+		"--", "set", "logical_switch_port", switchToRouterPrefix+nodeName, "type=router",
+		"options:router-port="+routerToSwitchPrefix+nodeName, "addresses="+"\""+nodeLRPMAC.String()+"\"")
 	if err != nil {
 		klog.Errorf("Failed to add logical port to switch, stdout: %q, stderr: %q, error: %v", stdout, stderr, err)
 		return err
@@ -567,7 +627,7 @@ func (oc *Controller) ensureNodeLogicalNetwork(nodeName string, hostSubnets []*n
 	}
 	stdout, stderr, err = util.RunOVNNbctl("set", "logical_switch", nodeName, "load_balancer="+oc.TCPLoadBalancerUUID)
 	if err != nil {
-		klog.Errorf("Failed to set logical switch %v's loadbalancer, stdout: %q, stderr: %q, error: %v", nodeName, stdout, stderr, err)
+		klog.Errorf("Failed to set logical switch %v's load balancer, stdout: %q, stderr: %q, error: %v", nodeName, stdout, stderr, err)
 		return err
 	}
 
@@ -585,7 +645,7 @@ func (oc *Controller) ensureNodeLogicalNetwork(nodeName string, hostSubnets []*n
 	}
 	stdout, stderr, err = util.RunOVNNbctl("add", "logical_switch", nodeName, "load_balancer", oc.UDPLoadBalancerUUID)
 	if err != nil {
-		klog.Errorf("Failed to add logical switch %v's loadbalancer, stdout: %q, stderr: %q, error: %v", nodeName, stdout, stderr, err)
+		klog.Errorf("Failed to add logical switch %v's load balancer, stdout: %q, stderr: %q, error: %v", nodeName, stdout, stderr, err)
 		return err
 	}
 
@@ -604,7 +664,7 @@ func (oc *Controller) ensureNodeLogicalNetwork(nodeName string, hostSubnets []*n
 		}
 		stdout, stderr, err = util.RunOVNNbctl("add", "logical_switch", nodeName, "load_balancer", oc.SCTPLoadBalancerUUID)
 		if err != nil {
-			klog.Errorf("Failed to add logical switch %v's loadbalancer, stdout: %q, stderr: %q, error: %v", nodeName, stdout, stderr, err)
+			klog.Errorf("Failed to add logical switch %v's load balancer, stdout: %q, stderr: %q, error: %v", nodeName, stdout, stderr, err)
 			return err
 		}
 
@@ -618,15 +678,7 @@ func (oc *Controller) ensureNodeLogicalNetwork(nodeName string, hostSubnets []*n
 		}
 	}
 	// Add the node to the logical switch cache
-	oc.lsMutex.Lock()
-	defer oc.lsMutex.Unlock()
-	if existing, ok := oc.logicalSwitchCache[nodeName]; ok && !reflect.DeepEqual(existing, hostSubnets) {
-		klog.Warningf("Node %q logical switch already in cache with subnet %s; replacing with %s", nodeName,
-			util.JoinIPNets(existing, ","), util.JoinIPNets(hostSubnets, ","))
-	}
-	oc.logicalSwitchCache[nodeName] = hostSubnets
-
-	return nil
+	return oc.lsManager.AddNode(nodeName, hostSubnets)
 }
 
 func (oc *Controller) addNodeAnnotations(node *kapi.Node, hostSubnets []*net.IPNet) error {
@@ -635,7 +687,18 @@ func (oc *Controller) addNodeAnnotations(node *kapi.Node, hostSubnets []*net.IPN
 		return fmt.Errorf("failed to marshal node %q annotation for subnet %s",
 			node.Name, util.JoinIPNets(hostSubnets, ","))
 	}
-	err = oc.kube.SetAnnotationsOnNode(node, nodeAnnotations)
+	// FIXME: the real solution is to reconcile the node object. Once we have a work-queue based
+	// implementation where we can add the item back to the work queue when it fails to
+	// reconcile, we can get rid of the PollImmediate.
+	err = utilwait.PollImmediate(OvnNodeAnnotationRetryInterval, OvnNodeAnnotationRetryTimeout, func() (bool, error) {
+		err = oc.kube.SetAnnotationsOnNode(node, nodeAnnotations)
+		if err != nil {
+			klog.Warningf("Failed to set node annotation, will retry for: %v",
+				OvnNodeAnnotationRetryTimeout)
+		}
+		return err == nil, nil
+	},
+	)
 	if err != nil {
 		return fmt.Errorf("failed to set node-subnets annotation on node %s: %v",
 			node.Name, err)
@@ -655,7 +718,7 @@ func (oc *Controller) addNode(node *kapi.Node) ([]*net.IPNet, error) {
 	// Node doesn't have a subnet assigned; reserve a new one for it
 	hostSubnets, err := oc.masterSubnetAllocator.AllocateNetworks()
 	if err != nil {
-		return nil, fmt.Errorf("Error allocating network for node %s: %v", node.Name, err)
+		return nil, fmt.Errorf("error allocating network for node %s: %v", node.Name, err)
 	}
 	klog.Infof("Allocated node %s HostSubnet %s", node.Name, util.JoinIPNets(hostSubnets, ","))
 
@@ -688,7 +751,7 @@ func (oc *Controller) addNode(node *kapi.Node) ([]*net.IPNet, error) {
 func (oc *Controller) deleteNodeHostSubnet(nodeName string, subnet *net.IPNet) error {
 	err := oc.masterSubnetAllocator.ReleaseNetwork(subnet)
 	if err != nil {
-		return fmt.Errorf("Error deleting subnet %v for node %q: %s", subnet, nodeName, err)
+		return fmt.Errorf("error deleting subnet %v for node %q: %s", subnet, nodeName, err)
 	}
 	klog.Infof("Deleted HostSubnet %v for node %s", subnet, nodeName)
 	return nil
@@ -697,20 +760,21 @@ func (oc *Controller) deleteNodeHostSubnet(nodeName string, subnet *net.IPNet) e
 func (oc *Controller) deleteNodeLogicalNetwork(nodeName string) error {
 	// Remove the logical switch associated with the node
 	if _, stderr, err := util.RunOVNNbctl("--if-exist", "ls-del", nodeName); err != nil {
-		return fmt.Errorf("Failed to delete logical switch %s, "+
+		return fmt.Errorf("failed to delete logical switch %s, "+
 			"stderr: %q, error: %v", nodeName, stderr, err)
 	}
 
 	// Remove the patch port that connects distributed router to node's logical switch
-	if _, stderr, err := util.RunOVNNbctl("--if-exist", "lrp-del", "rtos-"+nodeName); err != nil {
-		return fmt.Errorf("Failed to delete logical router port rtos-%s, "+
+	if _, stderr, err := util.RunOVNNbctl("--if-exist", "lrp-del", routerToSwitchPrefix+nodeName); err != nil {
+		return fmt.Errorf("failed to delete logical router port rtos-%s, "+
 			"stderr: %q, error: %v", nodeName, stderr, err)
 	}
 
 	return nil
 }
 
-func (oc *Controller) deleteNode(nodeName string, hostSubnets, joinSubnets []*net.IPNet) error {
+func (oc *Controller) deleteNode(nodeName string, hostSubnets, joinSubnets []*net.IPNet,
+	nodeLocalNatIPs []net.IP) error {
 	// Clean up as much as we can but don't hard error
 	for _, hostSubnet := range hostSubnets {
 		if err := oc.deleteNodeHostSubnet(nodeName, hostSubnet); err != nil {
@@ -722,13 +786,24 @@ func (oc *Controller) deleteNode(nodeName string, hostSubnets, joinSubnets []*ne
 			klog.Errorf("Error deleting node %s JoinSubnet %v: %v", nodeName, joinSubnet, err)
 		}
 	}
+	for _, nodeLocalNatIP := range nodeLocalNatIPs {
+		var err error
+		if utilnet.IsIPv6(nodeLocalNatIP) {
+			err = oc.nodeLocalNatIPv6Allocator.Release(nodeLocalNatIP)
+		} else {
+			err = oc.nodeLocalNatIPv4Allocator.Release(nodeLocalNatIP)
+		}
+		if err != nil {
+			klog.Errorf("Error deleting node %s's node local NAT IP %s from %v: %v", nodeName, nodeLocalNatIP, nodeLocalNatIPs, err)
+		}
+	}
 
 	if err := oc.deleteNodeLogicalNetwork(nodeName); err != nil {
 		klog.Errorf("Error deleting node %s logical network: %v", nodeName, err)
 	}
 
-	if err := gatewayCleanup(nodeName, hostSubnets); err != nil {
-		return fmt.Errorf("Failed to clean up node %s gateway: (%v)", nodeName, err)
+	if err := gatewayCleanup(nodeName); err != nil {
+		return fmt.Errorf("failed to clean up node %s gateway: (%v)", nodeName, err)
 	}
 
 	if err := oc.deleteNodeChassis(nodeName); err != nil {
@@ -784,24 +859,32 @@ func (oc *Controller) clearInitialNodeNetworkUnavailableCondition(origNode, newN
 		return err
 	})
 	if resultErr != nil {
-		klog.Errorf("status update failed for local node %s: %v", origNode.Name, resultErr)
+		klog.Errorf("Status update failed for local node %s: %v", origNode.Name, resultErr)
 	} else if cleared {
 		klog.Infof("Cleared node NetworkUnavailable/NoRouteCreated condition for %s", origNode.Name)
 	}
 }
 
 // delete chassis of the given nodeName/chassisName map
-func deleteChassis(chassisMap map[string]string) {
-	for nodeName, chassisName := range chassisMap {
+func deleteChassis(ovnSBClient goovn.Client, chassisMap map[string]string) {
+	cmds := make([]*goovn.OvnCommand, 0, len(chassisMap))
+	for chassisHostname, chassisName := range chassisMap {
 		if chassisName != "" {
-			klog.Infof("Deleting stale chassis %s (%s)", nodeName, chassisName)
-			_, stderr, err := util.RunOVNSbctl("--if-exist", "chassis-del", chassisName)
+			klog.Infof("Deleting stale chassis %s (%s)", chassisHostname, chassisName)
+			cmd, err := ovnSBClient.ChassisDel(chassisName)
 			if err != nil {
-				klog.Errorf("Failed to delete chassis with name %s for node %s: stderr: %s, error: %v",
-					chassisName, nodeName, stderr, err)
+				klog.Errorf("Unable to create the ChassisDel command for chassis: %s from the sbdb", chassisName)
+			} else {
+				cmds = append(cmds, cmd)
 			}
 		}
 	}
+
+	if len(cmds) != 0 {
+		if err := ovnSBClient.Execute(cmds...); err != nil {
+			klog.Errorf("Failed to delete chassis for node/chassis map %v: error: %v", chassisMap, err)
+		}
+	}
 }
 
 // this is the worker function that does the periodic sync of nodes from kube API
@@ -814,24 +897,21 @@ func (oc *Controller) syncNodesPeriodic() {
 		return
 	}
 
-	nodeNames := make([]string, len(nodes.Items))
+	nodeNames := make([]string, 0, len(nodes.Items))
 
 	for _, node := range nodes.Items {
 		nodeNames = append(nodeNames, node.Name)
 	}
 
-	chassisData, stderr, err := util.RunOVNSbctl("--data=bare", "--no-heading",
-		"--columns=name,hostname", "--format=json", "list", "Chassis")
+	chassisList, err := oc.ovnSBClient.ChassisList()
 	if err != nil {
-		klog.Errorf("Failed to get chassis list: stderr: %s, error: %v",
-			stderr, err)
+		klog.Errorf("Failed to get chassis list: error: %v", err)
 		return
 	}
 
-	chassisMap, err := oc.unmarshalChassisDataIntoMap([]byte(chassisData))
-	if err != nil {
-		klog.Errorf("Failed to unmarshal chassis data into chassis map, error: %v: %s", err, chassisData)
-		return
+	chassisMap := map[string]string{}
+	for _, chassis := range chassisList {
+		chassisMap[chassis.Hostname] = chassis.Name
 	}
 
 	//delete existing nodes from the chassis map.
@@ -839,7 +919,7 @@ func (oc *Controller) syncNodesPeriodic() {
 		delete(chassisMap, nodeName)
 	}
 
-	deleteChassis(chassisMap)
+	deleteChassis(oc.ovnSBClient, chassisMap)
 }
 
 func (oc *Controller) syncNodes(nodes []interface{}) {
@@ -857,25 +937,15 @@ func (oc *Controller) syncNodes(nodes []interface{}) {
 	// watchNodes() will be called for all existing nodes at startup anyway.
 	// Note that this list will include the 'join' cluster switch, which we
 	// do not want to delete.
-	var subnetAttr string
-	if config.IPv6Mode {
-		subnetAttr = "ipv6_prefix"
-	} else {
-		subnetAttr = "subnet"
-	}
-
-	chassisData, stderr, err := util.RunOVNSbctl("--data=bare", "--no-heading",
-		"--columns=name,hostname", "--format=json", "list", "Chassis")
+	chassisList, err := oc.ovnSBClient.ChassisList()
 	if err != nil {
-		klog.Errorf("Failed to get chassis list: stderr: %q, error: %v",
-			stderr, err)
+		klog.Errorf("Failed to get chassis list: error: %v", err)
 		return
 	}
 
-	chassisMap, err := oc.unmarshalChassisDataIntoMap([]byte(chassisData))
-	if err != nil {
-		klog.Errorf("Failed to unmarshal chassis data into chassis map, error: %v: %s", err, chassisData)
-		return
+	chassisMap := map[string]string{}
+	for _, chassis := range chassisList {
+		chassisMap[chassis.Hostname] = chassis.Name
 	}
 
 	//delete existing nodes from the chassis map.
@@ -884,8 +954,7 @@ func (oc *Controller) syncNodes(nodes []interface{}) {
 	}
 
 	nodeSwitches, stderr, err := util.RunOVNNbctl("--data=bare", "--no-heading",
-		"--columns=name,other-config", "find", "logical_switch",
-		"other-config:"+subnetAttr+"!=_")
+		"--columns=name,other-config", "find", "logical_switch")
 	if err != nil {
 		klog.Errorf("Failed to get node logical switches: stderr: %q, error: %v",
 			stderr, err)
@@ -929,6 +998,10 @@ func (oc *Controller) syncNodes(nodes []interface{}) {
 				subnets = append(subnets, subnet)
 			}
 		}
+		if len(subnets) == 0 {
+			continue
+		}
+
 		var tmp NodeSubnets
 		nodeSubnets, ok := NodeSubnetsMap[nodeName]
 		if !ok {
@@ -943,64 +1016,45 @@ func (oc *Controller) syncNodes(nodes []interface{}) {
 	}
 
 	for nodeName, nodeSubnets := range NodeSubnetsMap {
-		if err := oc.deleteNode(nodeName, nodeSubnets.hostSubnets, nodeSubnets.joinSubnets); err != nil {
+		if err := oc.deleteNode(nodeName, nodeSubnets.hostSubnets, nodeSubnets.joinSubnets, nil); err != nil {
 			klog.Error(err)
 		}
 		//remove the node from the chassis map so we don't delete it twice
 		delete(chassisMap, nodeName)
 	}
 
-	deleteChassis(chassisMap)
-}
-
-func (oc *Controller) unmarshalChassisDataIntoMap(chData []byte) (map[string]string, error) {
-	//map of node name to chassis name
-	chassisMap := make(map[string]string)
-
-	type chassisList struct {
-		Data     [][]string
-		Headings []string
-	}
-	var mapUnmarshal chassisList
-
-	if len(chData) == 0 {
-		return chassisMap, nil
-	}
-
-	err := json.Unmarshal(chData, &mapUnmarshal)
-
-	if err != nil {
-		return nil, fmt.Errorf("Error unmarshaling the chassis data: %s", err)
-	}
-
-	for _, chassis := range mapUnmarshal.Data {
-		if len(chassis) < 2 || chassis[0] == "" || chassis[1] == "" {
-			continue
-		}
-		chassisMap[chassis[1]] = chassis[0]
-	}
-
-	return chassisMap, nil
+	deleteChassis(oc.ovnSBClient, chassisMap)
 }
 
 func (oc *Controller) deleteNodeChassis(nodeName string) error {
-	chassisName, stderr, err := util.RunOVNSbctl("--data=bare", "--no-heading",
-		"--columns=name", "find", "Chassis",
-		"hostname="+nodeName)
+	var chNames []string
+
+	chassisList, err := oc.ovnSBClient.ChassisGet(nodeName)
 	if err != nil {
-		return fmt.Errorf("Failed to get chassis name for node %s: stderr: %q, error: %v",
-			nodeName, stderr, err)
+		return fmt.Errorf("failed to get chassis list for node %s: error: %v", nodeName, err)
 	}
 
-	if chassisName == "" {
-		klog.Warningf("Chassis name is empty for node: %s", nodeName)
-	} else {
-		_, stderr, err = util.RunOVNSbctl("--if-exist", "chassis-del", chassisName)
-		if err != nil {
-			return fmt.Errorf("Failed to delete chassis with name %s for node %s: stderr: %q, error: %v",
-				chassisName, nodeName, stderr, err)
+	cmds := make([]*goovn.OvnCommand, 0, len(chassisList))
+	for _, chassis := range chassisList {
+		if chassis.Name == "" {
+			klog.Warningf("Chassis name is empty for node: %s", nodeName)
+			continue
 		}
+		cmd, err := oc.ovnSBClient.ChassisDel(chassis.Name)
+		if err != nil {
+			return fmt.Errorf("unable to create the ChassisDel command for chassis: %s", chassis.Name)
+		}
+		chNames = append(chNames, chassis.Name)
+		cmds = append(cmds, cmd)
 	}
 
+	if len(cmds) == 0 {
+		return fmt.Errorf("failed to find chassis for node %s", nodeName)
+	}
+
+	if err = oc.ovnSBClient.Execute(cmds...); err != nil {
+		return fmt.Errorf("failed to delete chassis %q for node %s: error: %v",
+			strings.Join(chNames, ","), nodeName, err)
+	}
 	return nil
 }
