@@ -3,9 +3,10 @@ package util
 import (
 	"encoding/json"
 	"fmt"
+	"strings"
+
 	"k8s.io/client-go/tools/record"
 	"k8s.io/klog"
-	"strings"
 
 	kapi "k8s.io/api/core/v1"
 	"k8s.io/client-go/kubernetes"
@@ -15,13 +16,17 @@ import (
 	"k8s.io/client-go/tools/clientcmd"
 	"k8s.io/client-go/util/cert"
 
+	egressfirewallclientset "github.com/ovn-org/ovn-kubernetes/go-controller/pkg/crd/egressfirewall/v1/apis/clientset/versioned"
+	egressipclientset "github.com/ovn-org/ovn-kubernetes/go-controller/pkg/crd/egressip/v1/apis/clientset/versioned"
+	apiextensionsclientset "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
+
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/cni/types"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/config"
 )
 
-// NewClientset creates a Kubernetes clientset from either a kubeconfig,
+// NewClientsets creates a Kubernetes clientset from either a kubeconfig,
 // TLS properties, or an apiserver URL
-func NewClientset(conf *config.KubernetesConfig) (*kubernetes.Clientset, error) {
+func NewClientsets(conf *config.KubernetesConfig) (*kubernetes.Clientset, *egressipclientset.Clientset, *egressfirewallclientset.Clientset, *apiextensionsclientset.Clientset, error) {
 	var kconfig *rest.Config
 	var err error
 
@@ -29,8 +34,9 @@ func NewClientset(conf *config.KubernetesConfig) (*kubernetes.Clientset, error)
 		// uses the current context in kubeconfig
 		kconfig, err = clientcmd.BuildConfigFromFlags("", conf.Kubeconfig)
 	} else if strings.HasPrefix(conf.APIServer, "https") {
+		// TODO: Looks like the check conf.APIServer is redundant and can be removed
 		if conf.APIServer == "" || conf.Token == "" {
-			return nil, fmt.Errorf("TLS-secured apiservers require token and CA certificate")
+			return nil, nil, nil, nil, fmt.Errorf("TLS-secured apiservers require token and CA certificate")
 		}
 		kconfig = &rest.Config{
 			Host:        conf.APIServer,
@@ -38,7 +44,7 @@ func NewClientset(conf *config.KubernetesConfig) (*kubernetes.Clientset, error)
 		}
 		if conf.CACert != "" {
 			if _, err := cert.NewPool(conf.CACert); err != nil {
-				return nil, err
+				return nil, nil, nil, nil, err
 			}
 			kconfig.TLSClientConfig = rest.TLSClientConfig{CAFile: conf.CACert}
 		}
@@ -51,13 +57,34 @@ func NewClientset(conf *config.KubernetesConfig) (*kubernetes.Clientset, error)
 		kconfig, err = rest.InClusterConfig()
 	}
 	if err != nil {
-		return nil, err
+		return nil, nil, nil, nil, err
+	}
+
+	crdClientset, err := apiextensionsclientset.NewForConfig(kconfig)
+	if err != nil {
+		return nil, nil, nil, nil, err
+	}
+
+	// CRDS are not protobuf serializable, only JSON. So don't use that config for CRD clientsets
+	egressFirewallClientset, err := egressfirewallclientset.NewForConfig(kconfig)
+	if err != nil {
+		return nil, nil, nil, nil, err
+	}
+
+	egressIPClientset, err := egressipclientset.NewForConfig(kconfig)
+	if err != nil {
+		return nil, nil, nil, nil, err
 	}
 
 	kconfig.AcceptContentTypes = "application/vnd.kubernetes.protobuf,application/json"
 	kconfig.ContentType = "application/vnd.kubernetes.protobuf"
 
-	return kubernetes.NewForConfig(kconfig)
+	kClientset, err := kubernetes.NewForConfig(kconfig)
+	if err != nil {
+		return nil, nil, nil, nil, err
+	}
+
+	return kClientset, egressIPClientset, egressFirewallClientset, crdClientset, nil
 }
 
 // IsClusterIPSet checks if the service is an headless service or not
@@ -65,12 +92,20 @@ func IsClusterIPSet(service *kapi.Service) bool {
 	return service.Spec.ClusterIP != kapi.ClusterIPNone && service.Spec.ClusterIP != ""
 }
 
-// ValidateProtocol checks if the protocol is a valid kapi.Protocol type (TCP, UDP, or SCTP) or returns an error
-func ValidateProtocol(proto kapi.Protocol) (kapi.Protocol, error) {
-	if proto == kapi.ProtocolTCP || proto == kapi.ProtocolUDP || proto == kapi.ProtocolSCTP {
-		return proto, nil
+// ValidatePort checks if the port is non-zero and port protocol is valid
+func ValidatePort(proto kapi.Protocol, port int32) error {
+	if port <= 0 || port > 65535 {
+		return fmt.Errorf("invalid port number: %v", port)
 	}
-	return "", fmt.Errorf("protocol %s is not a valid protocol", proto)
+	return ValidateProtocol(proto)
+}
+
+// ValidateProtocol checks if the protocol is a valid kapi.Protocol type (TCP, UDP, or SCTP) or returns an error
+func ValidateProtocol(proto kapi.Protocol) error {
+	if proto == kapi.ProtocolTCP || proto == kapi.ProtocolUDP || proto == kapi.ProtocolSCTP {
+		return nil
+	}
+	return fmt.Errorf("protocol %s is not a valid protocol", proto)
 }
 
 // ServiceTypeHasClusterIP checks if the service has an associated ClusterIP or not
@@ -83,8 +118,8 @@ func ServiceTypeHasNodePort(service *kapi.Service) bool {
 	return service.Spec.Type == kapi.ServiceTypeNodePort || service.Spec.Type == kapi.ServiceTypeLoadBalancer
 }
 
-// GetNodeIP extracts the ip address from the node status in the  API
-func GetNodeIP(node *kapi.Node) (string, error) {
+// GetNodePrimaryIP extracts the primary IP address from the node status in the  API
+func GetNodePrimaryIP(node *kapi.Node) (string, error) {
 	for _, addr := range node.Status.Addresses {
 		if addr.Type == kapi.NodeInternalIP {
 			return addr.Address, nil
