@@ -4,7 +4,6 @@ import (
 	"bytes"
 	"encoding/json"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 	"runtime"
@@ -15,6 +14,7 @@ import (
 
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/config"
 	"github.com/prometheus/client_golang/prometheus"
+	"github.com/spf13/afero"
 
 	"k8s.io/klog"
 	kexec "k8s.io/utils/exec"
@@ -28,11 +28,14 @@ const (
 	ovsCommandTimeout  = 15
 	ovsVsctlCommand    = "ovs-vsctl"
 	ovsOfctlCommand    = "ovs-ofctl"
+	ovsDpctlCommand    = "ovs-dpctl"
 	ovsAppctlCommand   = "ovs-appctl"
 	ovnNbctlCommand    = "ovn-nbctl"
 	ovnSbctlCommand    = "ovn-sbctl"
 	ovnAppctlCommand   = "ovn-appctl"
 	ovsdbClientCommand = "ovsdb-client"
+	ovsdbToolCommand   = "ovsdb-tool"
+	arpingCommand      = "arping"
 	ipCommand          = "ip"
 	powershellCommand  = "powershell"
 	netshCommand       = "netsh"
@@ -44,8 +47,10 @@ const (
 )
 
 const (
-	nbdbCtlSock = "ovnnb_db.ctl"
-	sbdbCtlSock = "ovnsb_db.ctl"
+	nbdbCtlSock     = "ovnnb_db.ctl"
+	sbdbCtlSock     = "ovnsb_db.ctl"
+	OvnNbdbLocation = "/etc/ovn/ovnnb_db.db"
+	OvnSbdbLocation = "/etc/ovn/ovnsb_db.db"
 )
 
 var (
@@ -57,6 +62,9 @@ var (
 	savedOVNRunDir = ovnRunDir
 )
 
+var ovnCmdRetryCount = 200
+var AppFs = afero.NewOsFs()
+
 // PrepareTestConfig restores default config values. Used by testcases to
 // provide a pristine environment between tests.
 func PrepareTestConfig() {
@@ -72,7 +80,7 @@ func runningPlatform() (string, error) {
 	if runtime.GOOS == windowsOS {
 		return windowsOS, nil
 	}
-	fileContents, err := ioutil.ReadFile(osRelease)
+	fileContents, err := afero.ReadFile(AppFs, osRelease)
 	if err != nil {
 		return "", fmt.Errorf("failed to parse file %s (%v)", osRelease, err)
 	}
@@ -102,7 +110,7 @@ func runningPlatform() (string, error) {
 	} else if strings.Contains(platform, "VMware") {
 		return "Photon", nil
 	}
-	return "", fmt.Errorf("Unknown platform")
+	return "", fmt.Errorf("unknown platform")
 }
 
 // Exec runs various OVN and OVS utilities
@@ -110,14 +118,17 @@ type execHelper struct {
 	exec            kexec.Interface
 	ofctlPath       string
 	vsctlPath       string
+	dpctlPath       string
 	appctlPath      string
 	ovnappctlPath   string
 	nbctlPath       string
 	sbctlPath       string
 	ovnctlPath      string
 	ovsdbClientPath string
+	ovsdbToolPath   string
 	ovnRunDir       string
 	ipPath          string
+	arpingPath      string
 	powershellPath  string
 	netshPath       string
 	routePath       string
@@ -125,6 +136,44 @@ type execHelper struct {
 
 var runner *execHelper
 
+type ExecRunner interface {
+	RunCmd(cmd kexec.Cmd, cmdPath string, envVars []string, args ...string) (*bytes.Buffer, *bytes.Buffer, error)
+}
+
+// defaultExecRunner implements the methods defined in the ExecRunner interface
+type defaultExecRunner struct {
+}
+
+// RunCmd invokes the methods of the Cmd interfaces defined in k8s.io/utils/exec to execute commands
+// Note: the cmdPath and args parameter are used only for logging and is not processed
+func (runsvc *defaultExecRunner) RunCmd(cmd kexec.Cmd, cmdPath string, envVars []string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
+	if cmd == nil {
+		return &bytes.Buffer{}, &bytes.Buffer{}, fmt.Errorf("cmd object cannot be nil")
+	}
+	if len(envVars) != 0 {
+		cmd.SetEnv(envVars)
+	}
+	stdout := &bytes.Buffer{}
+	stderr := &bytes.Buffer{}
+
+	cmd.SetStdout(stdout)
+	cmd.SetStderr(stderr)
+
+	counter := atomic.AddUint64(&runCounter, 1)
+	logCmd := fmt.Sprintf("%s %s", cmdPath, strings.Join(args, " "))
+	klog.V(5).Infof("exec(%d): %s", counter, logCmd)
+
+	err := cmd.Run()
+	klog.V(5).Infof("exec(%d): stdout: %q", counter, stdout)
+	klog.V(5).Infof("exec(%d): stderr: %q", counter, stderr)
+	if err != nil {
+		klog.V(5).Infof("exec(%d): err: %v", counter, err)
+	}
+	return stdout, stderr, err
+}
+
+var runCmdExecRunner ExecRunner = &defaultExecRunner{}
+
 // SetExec validates executable paths and saves the given exec interface
 // to be used for running various OVS and OVN utilites
 func SetExec(exec kexec.Interface) error {
@@ -141,6 +190,10 @@ func SetExec(exec kexec.Interface) error {
 	if err != nil {
 		return err
 	}
+	runner.dpctlPath, err = exec.LookPath(ovsDpctlCommand)
+	if err != nil {
+		return err
+	}
 	runner.appctlPath, err = exec.LookPath(ovsAppctlCommand)
 	if err != nil {
 		return err
@@ -173,6 +226,10 @@ func SetExec(exec kexec.Interface) error {
 	if err != nil {
 		return err
 	}
+	runner.ovsdbToolPath, err = exec.LookPath(ovsdbToolCommand)
+	if err != nil {
+		return err
+	}
 
 	return nil
 }
@@ -201,6 +258,10 @@ func SetExecWithoutOVS(exec kexec.Interface) error {
 		if err != nil {
 			return err
 		}
+		runner.arpingPath, err = exec.LookPath(arpingCommand)
+		if err != nil {
+			return err
+		}
 	}
 	return nil
 }
@@ -235,34 +296,17 @@ func GetExec() kexec.Interface {
 var runCounter uint64
 
 func runCmd(cmd kexec.Cmd, cmdPath string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
-	stdout := &bytes.Buffer{}
-	stderr := &bytes.Buffer{}
-
-	cmd.SetStdout(stdout)
-	cmd.SetStderr(stderr)
-
-	counter := atomic.AddUint64(&runCounter, 1)
-	logCmd := fmt.Sprintf("%s %s", cmdPath, strings.Join(args, " "))
-	klog.V(5).Infof("exec(%d): %s", counter, logCmd)
-
-	err := cmd.Run()
-	klog.V(5).Infof("exec(%d): stdout: %q", counter, stdout)
-	klog.V(5).Infof("exec(%d): stderr: %q", counter, stderr)
-	if err != nil {
-		klog.V(5).Infof("exec(%d): err: %v", counter, err)
-	}
-	return stdout, stderr, err
+	return runCmdExecRunner.RunCmd(cmd, cmdPath, []string{}, args...)
 }
 
 func run(cmdPath string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
 	cmd := runner.exec.Command(cmdPath, args...)
-	return runCmd(cmd, cmdPath, args...)
+	return runCmdExecRunner.RunCmd(cmd, cmdPath, []string{}, args...)
 }
 
 func runWithEnvVars(cmdPath string, envVars []string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
 	cmd := runner.exec.Command(cmdPath, args...)
-	cmd.SetEnv(envVars)
-	return runCmd(cmd, cmdPath, args...)
+	return runCmdExecRunner.RunCmd(cmd, cmdPath, envVars, args...)
 }
 
 // RunOVSOfctl runs a command via ovs-ofctl.
@@ -271,6 +315,12 @@ func RunOVSOfctl(args ...string) (string, string, error) {
 	return strings.Trim(stdout.String(), "\" \n"), stderr.String(), err
 }
 
+// RunOVSDpctl runs a command via ovs-dpctl.
+func RunOVSDpctl(args ...string) (string, string, error) {
+	stdout, stderr, err := run(runner.dpctlPath, args...)
+	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
+}
+
 // RunOVSVsctl runs a command via ovs-vsctl.
 func RunOVSVsctl(args ...string) (string, string, error) {
 	cmdArgs := []string{fmt.Sprintf("--timeout=%d", ovsCommandTimeout)}
@@ -305,7 +355,7 @@ func RunOVNAppctlWithTimeout(timeout int, args ...string) (string, string, error
 // poll waitng for service to become available
 func runOVNretry(cmdPath string, envVars []string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
 
-	retriesLeft := 200
+	retriesLeft := ovnCmdRetryCount
 	for {
 		stdout, stderr, err := runWithEnvVars(cmdPath, envVars, args...)
 		if err == nil {
@@ -335,7 +385,7 @@ var SkippedNbctlDaemonCounter uint64
 func getNbctlSocketPath() (string, error) {
 	// Try already-set OVN_NB_DAEMON environment variable
 	if nbctlSocketPath := os.Getenv("OVN_NB_DAEMON"); nbctlSocketPath != "" {
-		if _, err := os.Stat(nbctlSocketPath); err != nil {
+		if _, err := AppFs.Stat(nbctlSocketPath); err != nil {
 			return "", fmt.Errorf("OVN_NB_DAEMON ovn-nbctl daemon control socket %s missing: %v",
 				nbctlSocketPath, err)
 		}
@@ -348,10 +398,10 @@ func getNbctlSocketPath() (string, error) {
 	for _, runDir := range dirs {
 		// Try autodetecting the socket path based on the nbctl daemon pid
 		pidfile := filepath.Join(runDir, "ovn-nbctl.pid")
-		if pid, err := ioutil.ReadFile(pidfile); err == nil {
+		if pid, err := afero.ReadFile(AppFs, pidfile); err == nil {
 			fname := fmt.Sprintf("ovn-nbctl.%s.ctl", strings.TrimSpace(string(pid)))
 			nbctlSocketPath := filepath.Join(runDir, fname)
-			if _, err := os.Stat(nbctlSocketPath); err == nil {
+			if _, err := AppFs.Stat(nbctlSocketPath); err == nil {
 				return "OVN_NB_DAEMON=" + nbctlSocketPath, nil
 			}
 		}
@@ -475,6 +525,12 @@ func RunOVSDBClient(args ...string) (string, string, error) {
 	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
 }
 
+// RunOVSDBTool runs an 'ovsdb-tool [OPTIONS] COMMAND [ARG...] command'.
+func RunOVSDBTool(args ...string) (string, string, error) {
+	stdout, stderr, err := run(runner.ovsdbToolPath, args...)
+	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
+}
+
 // RunOVSDBClientOVN runs an 'ovsdb-client [OPTIONS] COMMAND [SERVER] [ARG...] command' against OVN NB database.
 func RunOVSDBClientOVNNB(command string, args ...string) (string, string, error) {
 	cmdArgs := getNbOVSDBArgs(command, args...)
@@ -518,10 +574,11 @@ func RunOVNSBAppCtl(args ...string) (string, string, error) {
 }
 
 // RunOVNNorthAppCtl runs an 'ovs-appctl -t ovn-northd command'.
+// TODO: Currently no module is invoking this function, will need to consider adding an unit test when actively used
 func RunOVNNorthAppCtl(args ...string) (string, string, error) {
 	var cmdArgs []string
 
-	pid, err := ioutil.ReadFile(runner.ovnRunDir + "ovn-northd.pid")
+	pid, err := afero.ReadFile(AppFs, runner.ovnRunDir+"ovn-northd.pid")
 	if err != nil {
 		return "", "", fmt.Errorf("failed to run the command since failed to get ovn-northd's pid: %v", err)
 	}
@@ -535,12 +592,50 @@ func RunOVNNorthAppCtl(args ...string) (string, string, error) {
 	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
 }
 
+// RunOVNControllerAppCtl runs an 'ovs-appctl -t ovn-controller.pid.ctl command'.
+func RunOVNControllerAppCtl(args ...string) (string, string, error) {
+	var cmdArgs []string
+	pid, err := afero.ReadFile(AppFs, runner.ovnRunDir+"ovn-controller.pid")
+	if err != nil {
+		return "", "", fmt.Errorf("failed to get ovn-controller pid : %v", err)
+	}
+	cmdArgs = []string{
+		"-t",
+		runner.ovnRunDir + fmt.Sprintf("ovn-controller.%s.ctl", strings.TrimSpace(string(pid))),
+	}
+	cmdArgs = append(cmdArgs, args...)
+	stdout, stderr, err := runOVNretry(runner.ovnappctlPath, nil, cmdArgs...)
+	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
+}
+
+// RunOvsVswitchdAppCtl runs an 'ovs-appctl -t /var/run/openvsiwthc/ovs-vswitchd.pid.ctl command'
+func RunOvsVswitchdAppCtl(args ...string) (string, string, error) {
+	var cmdArgs []string
+	pid, err := afero.ReadFile(AppFs, savedOVSRunDir+"ovs-vswitchd.pid")
+	if err != nil {
+		return "", "", fmt.Errorf("failed to get ovs-vswitch pid : %v", err)
+	}
+	cmdArgs = []string{
+		"-t",
+		savedOVSRunDir + fmt.Sprintf("ovs-vswitchd.%s.ctl", strings.TrimSpace(string(pid))),
+	}
+	cmdArgs = append(cmdArgs, args...)
+	stdout, stderr, err := runOVNretry(runner.appctlPath, nil, cmdArgs...)
+	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
+}
+
 // RunIP runs a command via the iproute2 "ip" utility
 func RunIP(args ...string) (string, string, error) {
 	stdout, stderr, err := run(runner.ipPath, args...)
 	return strings.TrimSpace(stdout.String()), stderr.String(), err
 }
 
+// RunArping runs a command via the "arping" utility
+func RunArping(args ...string) (string, string, error) {
+	stdout, stderr, err := run(runner.arpingPath, args...)
+	return strings.TrimSpace(stdout.String()), stderr.String(), err
+}
+
 // RunPowershell runs a command via the Windows powershell utility
 func RunPowershell(args ...string) (string, string, error) {
 	stdout, stderr, err := run(runner.powershellPath, args...)
@@ -559,27 +654,12 @@ func RunRoute(args ...string) (string, string, error) {
 	return strings.TrimSpace(stdout.String()), stderr.String(), err
 }
 
-// RawExec runs the given command via the exec interface. Should only be used
-// for early calls before configuration is read.
-func RawExec(cmdPath string, args ...string) (string, string, error) {
-	// If the command is not a path to a binary, try finding it
-	if filepath.Base(cmdPath) == cmdPath {
-		var err error
-		cmdPath, err = runner.exec.LookPath(cmdPath)
-		if err != nil {
-			return "", "", err
-		}
-	}
-	stdout, stderr, err := run(cmdPath, args...)
-	return strings.TrimSpace(stdout.String()), stderr.String(), err
-}
-
-// AddNormalActionOFFlow replaces flows in the bridge with a NORMAL action flow
-func AddNormalActionOFFlow(bridgeName string) (string, string, error) {
+// AddFloodActionOFFlow replaces flows in the bridge with a FLOOD action flow
+func AddFloodActionOFFlow(bridgeName string) (string, string, error) {
 	args := []string{"-O", "OpenFlow13", "replace-flows", bridgeName, "-"}
 
 	stdin := &bytes.Buffer{}
-	stdin.Write([]byte("table=0,priority=0,actions=NORMAL\n"))
+	stdin.Write([]byte("table=0,priority=0,actions=FLOOD\n"))
 
 	cmd := runner.exec.Command(runner.ofctlPath, args...)
 	cmd.SetStdin(stdin)
